/*
 * Clustions: clustered notions - notion graining
 * (C) Martin Saturka, http://enduce.tangloid.net/
 * Version 0.92, released under the MIT license.
**/

#include "clustions.h"
#include <cmath>

namespace Enduce {
  static Clustions clu_static;
  Clustions *clu = &clu_static;
}

Enduce::Clustions::Clustions(void) {
  this->m_thres = 0;
  this->m_maxcycl = Enduce::DEFMAXCYCL;
}

Enduce::Clustions::Clustions(double thres) {
  this->m_thres = 0;
  this->m_maxcycl = Enduce::DEFMAXCYCL;

  if ((0.0 < thres) && (1.0 >= thres)) {
    this->m_thres = thres;
  }
}

bool Enduce::Clustions::thresholds(double thres) {
  if ((0.0 > thres) || (1.0 < thres)) {
    return false;
  }

  this->m_thres = thres;
  return true;
}

int Enduce::Clustions::thresholds(int thlen, double *thres) {
  if (0 >= thlen) {
    return 0;
  }

  thres[0] = this->m_thres;
  return 1;
}

bool Enduce::Clustions::setLimits(int maxcycl) {
  if (0 >= maxcycl) {
    return false;
  }

  this->m_maxcycl = maxcycl;
  return true;
}

int Enduce::Clustions::getLimits(int limlen, int *limits) {
  if (0 >= limlen) {
    return 0;
  }

  limits[0] = this->m_maxcycl;
  return 1;
}

bool Enduce::Clustions::cluster(double *centers, double *datamatrix, int *sizes,
  int *grains, double *simils, int *info) {

  int cencount = sizes[0];
  int veccount = sizes[1];
  int expcount = sizes[2];

  if ((0 == cencount) || (0 == veccount) || (0 == expcount)) {
    return false;
  }

  int *censizes = new int [cencount];

  for (int ivec = 0; ivec < veccount; ivec++) {
    simils[ivec] = -1.0;
    grains[ivec] = -1;
  }

  double *curvector;
  double *curcenter;
  double cursimil;

  bool tocycle = true;
  bool modified = true;
  int totcyc = 0;
  int oldgrain;
  while (tocycle) {
    totcyc += 1;
    modified = false;
    curvector = datamatrix;

    for (int ivec = 0; ivec < veccount; ivec++, curvector += expcount) {
      oldgrain = grains[ivec];

      if (-1 < oldgrain) {
        curcenter = centers + (oldgrain * expcount);
        cursimil = this->evalSimil(curvector, curcenter, expcount);
        simils[ivec] = cursimil;
      }
      else {
        modified = true;
      }

      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (icen != oldgrain) {
          cursimil = this->evalSimil(curvector, curcenter, expcount);
          if (cursimil > simils[ivec]) {
            simils[ivec] = cursimil;
            grains[ivec] = icen;
            modified = true;
          }
        }

      }
    }

    if ((!modified) || (totcyc == this->m_maxcycl)) {
      tocycle = false;
    }

    if (modified) {
      //cycle centers: zero center sizes;
      for (int icen = 0; icen < cencount; icen++) {
        censizes[icen] = 0;
      }
      //cycle vectors: add to center sizes;
      for (int ivec = 0; ivec < veccount; ivec++) {
        if (-1 != grains[ivec]) {
          censizes[grains[ivec]] += 1;
        }
      }
      //cycle centers: zero non-empty centers;
      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (0 == censizes[icen]) {
          continue;
        }
        for (int ipos = 0; ipos < expcount; ipos++) {
          curcenter[ipos] = 0.0;
        }
      }
      //cycle vectors: add to centers;
      curvector = datamatrix;
      for (int ivec = 0; ivec < veccount; ivec++, curvector += expcount) {
        if (-1 != grains[ivec]) {
          curcenter = centers + (grains[ivec] * expcount);
          for (int ipos = 0; ipos < expcount; ipos++) {
            curcenter[ipos] += curvector[ipos];
          }
        }
      }
      //cycle centers: divide centers by their sizes;
      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (0 == censizes[icen]) {
          continue;
        }
        for (int ipos = 0; ipos < expcount; ipos++) {
          curcenter[ipos] /= censizes[icen];
        }
      }
    }

  }

  int converged = 1;
  if (modified) {
    converged = 0;
  }
  info[0] = converged;
  info[1] = totcyc;

  delete [] censizes;

  return true;
}

bool Enduce::Clustions::cluster(double *centers, double *datamatrix, int *sizes,
  int *grainsneg, int *grainspos, double *similsneg, double *similspos, int *info) {

  int cencount = sizes[0];
  int veccount = sizes[1];
  int expcount = sizes[2];

  if ((0 == cencount) || (0 == veccount) || (0 == expcount)) {
    return false;
  }

  int *censizes = new int [cencount];

  double curvalue;
  double *vectneg = new double [expcount];
  double *vectpos = new double [expcount];

  for (int ivec = 0; ivec < veccount; ivec++) {
    similsneg[ivec] = -1.0;
    grainsneg[ivec] = -1;
    similspos[ivec] = -1.0;
    grainspos[ivec] = -1;
  }

  double *curvector;
  double *curcenter;
  double cursimil;

  bool tocycle = true;
  bool modified = true;
  int totcyc = 0;
  int oldgrainneg;
  int oldgrainpos;
  while (tocycle) {
    totcyc += 1;
    modified = false;
    curvector = datamatrix;

    for (int ivec = 0; ivec < veccount; ivec++, curvector += expcount) {

      for (int ipos = 0; ipos < expcount; ipos++) {
        curvalue = curvector[ipos];
        if (0 < curvalue) {
          vectneg[ipos] = 0.0;
          vectpos[ipos] = curvalue;
        }
        else if (0 > curvalue) {
          vectneg[ipos] = fabs(curvalue);
          vectpos[ipos] = 0.0;
        }
        else {
          vectneg[ipos] = 0.0;
          vectpos[ipos] = 0.0;
        }
      }

      oldgrainneg = grainsneg[ivec];
      oldgrainpos = grainspos[ivec];

      if (-1 < oldgrainneg) {
        curcenter = centers + (oldgrainneg * expcount);
        cursimil = this->evalSimil(vectneg, curcenter, expcount);
        similsneg[ivec] = cursimil;
      }
      else {
        modified = true;
      }
      if (-1 < oldgrainpos) {
        curcenter = centers + (oldgrainpos * expcount);
        cursimil = this->evalSimil(vectpos, curcenter, expcount);
        similspos[ivec] = cursimil;
      }
      else {
        modified = true;
      }

      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (icen != oldgrainneg) {
          cursimil = this->evalSimil(vectneg, curcenter, expcount);
          if (cursimil > similsneg[ivec]) {
            similsneg[ivec] = cursimil;
            grainsneg[ivec] = icen;
            modified = true;
          }
        }
        if (icen != oldgrainpos) {
          cursimil = this->evalSimil(vectpos, curcenter, expcount);
          if (cursimil > similspos[ivec]) {
            similspos[ivec] = cursimil;
            grainspos[ivec] = icen;
            modified = true;
          }
        }

      }
    }

    if ((!modified) || (totcyc == this->m_maxcycl)) {
      tocycle = false;
    }

    if (modified) {
      //cycle centers: zero center sizes;
      for (int icen = 0; icen < cencount; icen++) {
        censizes[icen] = 0;
      }
      //cycle vectors: add to center sizes;
      for (int ivec = 0; ivec < veccount; ivec++) {
        if (-1 != grainsneg[ivec]) {
          censizes[grainsneg[ivec]] += 1;
        }
        if (-1 != grainspos[ivec]) {
          censizes[grainspos[ivec]] += 1;
        }
      }
      //cycle centers: zero non-empty centers;
      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (0 == censizes[icen]) {
          continue;
        }
        for (int ipos = 0; ipos < expcount; ipos++) {
          curcenter[ipos] = 0.0;
        }
      }
      //cycle vectors: add to centers;
      curvector = datamatrix;
      for (int ivec = 0; ivec < veccount; ivec++, curvector += expcount) {
        if (-1 != grainsneg[ivec]) {
          curcenter = centers + (grainsneg[ivec] * expcount);
          for (int ipos = 0; ipos < expcount; ipos++) {
            if (0 > curvector[ipos]) {
              curcenter[ipos] += fabs(curvector[ipos]);
            }
          }
        }
        if (-1 != grainspos[ivec]) {
          curcenter = centers + (grainspos[ivec] * expcount);
          for (int ipos = 0; ipos < expcount; ipos++) {
            if (0 < curvector[ipos]) {
              curcenter[ipos] += curvector[ipos];
            }
          }
        }
      }
      //cycle centers: divide centers by their sizes;
      curcenter = centers;
      for (int icen = 0; icen < cencount; icen++, curcenter += expcount) {
        if (0 == censizes[icen]) {
          continue;
        }
        for (int ipos = 0; ipos < expcount; ipos++) {
          curcenter[ipos] /= censizes[icen];
        }
      }
    }

  }

  int converged = 1;
  if (modified) {
    converged = 0;
  }
  info[0] = converged;
  info[1] = totcyc;

  delete [] censizes;
  delete [] vectneg;
  delete [] vectpos;

  return true;
}


double Enduce::Clustions::evalSimil(double *curvector, double *curcenter, int expcount) {
  double cursimil = 0.0;
  double curnumer = 0.0;
  double curdenom = 0.0;

  double curvalue;
  double curgrain;

  for (int ipos = 0; ipos < expcount; ipos++) {
    curvalue = curvector[ipos];
    curgrain = curcenter[ipos];
    if ((this->m_thres > curvalue) && (this->m_thres > curgrain)) {
      curvalue = 0.0;
      curgrain = 0.0;
    }
    if (0.0 > curvalue) {
      curvalue = 0.0;
    }
    if (0.0 > curgrain) {
      curgrain = 0.0;
    }

    if (curvalue <= curgrain) {
      curnumer += curvalue;
      curdenom += curgrain;
    }
    else {
      curnumer += curgrain;
      curdenom += curvalue;
    }
  }
  if (0.0 < curdenom) {
    cursimil = curnumer / curdenom;
  }

  return cursimil;
}

