/*
 * Contifiers: natural aggregation of interval valued pairs.
 * (C) Martin Saturka, http://enduce.tangloid.net/
 * Version 0.92, released under the MIT license.
**/

#include "contifiers.h"

namespace Enduce {
  static Contifiers con_static;
  Contifiers *con = &con_static;
}

Enduce::Contifiers::Contifiers(void) {
  this->m_thragg = 0.0;

  this->m_aggdirc = 1;
  this->m_aggtype = 1;
}

Enduce::Contifiers::Contifiers(double thragg) {
  this->m_thragg = 0.0;
  if ((0.0 < thragg) && (1.0 > thragg)) {
    this->m_thragg = thragg;
  }

  this->m_aggdirc = 1;
  this->m_aggtype = 1;
}

bool Enduce::Contifiers::thresholds(double thres) {
  if ((0.0 > thres) || (1.0 < thres)) {
      return false;
  }

  this->m_thragg = thres;
  return true;
}

int Enduce::Contifiers::thresholds(int thlen, double *thres) {
  if (!thres) {
    return 0;
  }
  if (0 == thlen) {
    return 0;
  }

  thres[0] = this->m_thragg;
  return 1;
}

bool Enduce::Contifiers::setModes(int aggdirc, int aggtype) {
  int iaggdirc = -1;
  int iaggtype = -1;

  for (int itype = 0; itype < Enduce::AGGDIRCSCNT; itype++) {
    if (itype == aggdirc) {
      iaggdirc = itype;
    }
  }
  for (int itype = 0; itype < Enduce::AGGTYPESCNT; itype++) {
    if (itype == aggtype) {
      iaggtype = itype;
    }
  }

  if ((-1 == iaggdirc) || (-1 == iaggtype)) {
    return false;
  }

  this->m_aggdirc = iaggdirc;
  this->m_aggtype = iaggtype;

  return true;
}

int Enduce::Contifiers::getModes(int mdlen, int *modes) {
  if (!modes) {
    return 0;
  }
  if (2 > mdlen) {
    return 0;
  }

  modes[0] = this->m_aggdirc;
  modes[1] = this->m_aggtype;

  return 2;
}

bool Enduce::Contifiers::aggregate(int vallen, double *valuesx, double *valuesy, double *results) {
  bool retval;
  double *antc;
  double *succ;

  if ((!valuesx) || (!valuesy) || (!results)) {
    return false;
  }
  results[0] = 0.0;

  if (0 == this->m_aggdirc) {
    if (2 == this->m_aggtype) {
      retval = aggMutType2(vallen, valuesx, valuesy, results);
    }
    else if (1 == this->m_aggtype) {
      retval = aggMutType1(vallen, valuesx, valuesy, results);
    }
    else {
      retval = aggMutType0(vallen, valuesx, valuesy, results);
    }
  }
  else {
    if (1 == this->m_aggdirc) {
      antc = valuesx;
      succ = valuesy;
    }
    else {
      antc = valuesy;
      succ = valuesx;
    }

    if (2 == this->m_aggtype) {
      retval = aggDirType2(vallen, antc, succ, results);
    }
    else if (1 == this->m_aggtype) {
      retval = aggDirType1(vallen, antc, succ, results);
    }
    else {
      retval = aggDirType0(vallen, antc, succ, results);
    }
  }

  return retval;
}

bool Enduce::Contifiers::aggMutType0(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double sync1, sync2;

  for (int ind = 0; ind < vallen; ind++) {
    sync1 = valuesx[ind];
    sync2 = valuesy[ind];

    if ((0.0 > sync1) || (1.0 < sync1) || (0.0 > sync2) || (1.0 < sync2)) {
      return false;
    }

    if (sync1 <= sync2) {
      if (thrpos <= sync2) {
        denom += sync2;
        numer += sync2 * sync1;
      }
    }
    else {
      if (thrpos <= sync1) {
        denom += sync1;
        numer += sync1 * sync2;
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}

bool Enduce::Contifiers::aggMutType1(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double sync1, sync2;

  for (int ind = 0; ind < vallen; ind++) {
    sync1 = valuesx[ind];
    sync2 = valuesy[ind];

    if ((0.0 > sync1) || (1.0 < sync1) || (0.0 > sync2) || (1.0 < sync2)) {
      return false;
    }

    if (sync1 <= sync2) {
      if (thrpos <= sync2) {
        denom += sync2;
        numer += sync1;
      }
    }
    else {
      if (thrpos <= sync1) {
        denom += sync1;
        numer += sync2;
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}

bool Enduce::Contifiers::aggMutType2(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double sync1, sync2;

  for (int ind = 0; ind < vallen; ind++) {
    sync1 = valuesx[ind];
    sync2 = valuesy[ind];

    if ((0.0 > sync1) || (1.0 < sync1) || (0.0 > sync2) || (1.0 < sync2)) {
      return false;
    }

    if (sync1 <= sync2) {
      if (thrpos <= sync2) {
        denom += sync2;
        numer += sync2 * (1 + sync1 - sync2);
      }
    }
    else {
      if (thrpos <= sync1) {
        denom += sync1;
        numer += sync1 * (1 + sync2 - sync1);
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}

bool Enduce::Contifiers::aggDirType0(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double antc, succ;

  for (int ind = 0; ind < vallen; ind++) {
    antc = valuesx[ind];
    succ = valuesy[ind];

    if ((0.0 > antc) || (1.0 < antc) || (0.0 > succ) || (1.0 < succ)) {
      return false;
    }

    if (thrpos <= antc) {
      denom += antc;
      if (antc <= succ) {
        numer += antc;
      }
      else {
        numer += antc * succ;
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}

bool Enduce::Contifiers::aggDirType1(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double antc, succ;

  for (int ind = 0; ind < vallen; ind++) {
    antc = valuesx[ind];
    succ = valuesy[ind];

    if ((0.0 > antc) || (1.0 < antc) || (0.0 > succ) || (1.0 < succ)) {
      return false;
    }

    if (thrpos <= antc) {
      denom += antc;
      if (antc <= succ) {
        numer += antc;
      }
      else {
        numer += succ;
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}

bool Enduce::Contifiers::aggDirType2(int vallen, double *valuesx, double *valuesy, double *results) {
  double thrpos = this->m_thragg;
  double numer = 0.0, denom = 0.0;
  double antc, succ;

  for (int ind = 0; ind < vallen; ind++) {
    antc = valuesx[ind];
    succ = valuesy[ind];

    if ((0.0 > antc) || (1.0 < antc) || (0.0 > succ) || (1.0 < succ)) {
      return false;
    }

    if (thrpos <= antc) {
      denom += antc;
      if (antc <= succ) {
        numer += antc;
      }
      else {
        numer += antc * (1 + succ - antc);
      }
    }

  }

  if (0.0 < denom) {
    results[0] = numer / denom;
  }
  else {
    results[0] = 0.0;
  }

  return true;
}



