/*
 * Dinorms: natural aggregation of dichotomic values.
 * (C) Martin Saturka, http://enduce.tangloid.net/
 * Version 0.92, released under the MIT license.
**/

#include "dinorms.h"

namespace Enduce {
  static Dinorms din_static;
  Dinorms *din = &din_static;
}

Enduce::Dinorms::Dinorms(void) {
  this->m_thragg = 0.0;
  this->m_thrcmb = 0.0;

  this->m_aggtype = 0;
  this->m_cmbtype = 0;
}

Enduce::Dinorms::Dinorms(double thragg, double thrcmb) {
  this->m_thragg = 0.0;
  if ((0.0 <= thragg) && (1.0 >= thragg)) {
    this->m_thragg = thragg;
  }
  this->m_thrcmb = 0.0;
  if ((0.0 <= thrcmb) && (1.0 >= thrcmb)) {
    this->m_thrcmb = thrcmb;
  }

  this->m_aggtype = 0;
  this->m_cmbtype = 0;
}

bool Enduce::Dinorms::thresholds(double thragg, double thrcmb) {
  if ((0.0 > thragg) || (1.0 < thragg)) {
    return false;
  }
  if ((0.0 > thrcmb) || (1.0 < thrcmb)) {
    return false;
  }

  this->m_thragg = thragg;
  this->m_thrcmb = thrcmb;
  return true;
}

int Enduce::Dinorms::thresholds(int thlen, double *thres) {
  if (!thres) {
    return 0;
  }
  if (2 > thlen) {
    return 0;
  }

  thres[0] = this->m_thragg;
  thres[1] = this->m_thrcmb;

  return 2;
}

bool Enduce::Dinorms::setModes(int aggtype, int cmbtype) {
  int iaggtype = -1;
  int icmbtype = -1;

  for (int itype = 0; itype < AGGTYPESCNT; itype++) {
    if (itype == aggtype) {
      iaggtype = itype;
    }
  }
  for (int itype = 0; itype < CMBTYPESCNT; itype++) {
    if (itype == cmbtype) {
      icmbtype = itype;
    }
  }

  if ((-1 == iaggtype) || (-1 == icmbtype)) {
    return false;
  }

  this->m_aggtype = iaggtype;
  this->m_cmbtype = icmbtype;
  return true;
}

int Enduce::Dinorms::getModes(int mdlen, int *modes) {
  if (!modes) {
    return 0;
  }
  if (2 > mdlen) {
    return 0;
  }

  modes[0] = this->m_aggtype;
  modes[1] = this->m_cmbtype;

  return 2;
}

bool Enduce::Dinorms::aggregate(int vallen, double *values, double *results) {
  bool retval;

  if ((!values) || (!results)) {
    return false;
  }
  for (int ind = 0; ind < 3; ind++) {
    results[ind] = 0.0;
  }

  if (2 == this->m_aggtype) {
    retval = aggType2(vallen, values, results);
  }
  else if (1 == this->m_aggtype) {
    retval = aggType1(vallen, values, results);
  }
  else {
    retval = aggType0(vallen, values, results);
  }
  if (!retval) {
    return false;
  }

  if (3 == this->m_cmbtype) {
    retval = cmbType3(results);
  }
  else if (2 == this->m_cmbtype) {
    retval = cmbType2(results);
  }
  else if (1 == this->m_cmbtype) {
    retval = cmbType1(results);
  }
  else {
    retval = cmbType0(results);
  }

  return retval;
}

bool Enduce::Dinorms::aggType0(int vallen, double *values, double *results) {
  double thrpos = this->m_thragg;
  double thrneg = -1.0 * this->m_thragg;
  double vneg = 0.0, vpos = 0.0;
  double curvalue;

  for (int ind = 0; ind < vallen; ind++) {
    curvalue = values[ind];
    if (0.0 == curvalue) {
      continue;
    }

    if (thrneg >= curvalue) {
      if (-1.0 > curvalue) {
        return false;
      }
      vneg = (vneg < curvalue) ? vneg : curvalue;
    }
    else if (thrpos <= curvalue) {
      if (1.0 < curvalue) {
        return false;
      }
      vpos = (vpos > curvalue) ? vpos : curvalue;
    }
  }

  results[0] = vneg;
  results[1] = vpos;
  return true;
}

bool Enduce::Dinorms::aggType1(int vallen, double *values, double *results) {
  double thrpos = this->m_thragg;
  double thrneg = -1.0 * this->m_thragg;
  double vneg = 0.0, vpos = 0.0;
  double curvalue;

  for (int ind = 0; ind < vallen; ind++) {
    curvalue = values[ind];
    if (0.0 == curvalue) {
      continue;
    }

    if (thrneg >= curvalue) {
      if (-1.0 > curvalue) {
        return false;
      }
      vneg += curvalue + (vneg * curvalue);
    }
    else if (thrpos <= curvalue) {
      if (1.0 < curvalue) {
        return false;
      }
      vpos += curvalue - (vpos * curvalue);
    }
  }

  results[0] = vneg;
  results[1] = vpos;
  return true;
}

bool Enduce::Dinorms::aggType2(int vallen, double *values, double *results) {
  double thrpos = this->m_thragg;
  double thrneg = -1.0 * this->m_thragg;
  double vneg = 0.0, vpos = 0.0;
  double curvalue;
  double vaux;

  for (int ind = 0; ind < vallen; ind++) {
    curvalue = values[ind];
    if (0.0 == curvalue) {
      continue;
    }

    if (thrneg >= curvalue) {
      if (-1.0 > curvalue) {
        return false;
      }
      vaux = vneg + curvalue;
      vneg = (-1.0 < vaux) ? vaux : -1.0;
    }
    else if (thrpos <= curvalue) {
      if (1.0 < curvalue) {
        return false;
      }
      vaux = vpos + curvalue;
      vpos = (1.0 > vaux) ? vaux : 1.0;
    }
  }

  results[0] = vneg;
  results[1] = vpos;
  return true;
}

bool Enduce::Dinorms::cmbType0(double *combs) {
  if (!combs) {
    return false;
  }
  double vneg = combs[0];
  double vpos = combs[1];
  if ((-1.0 > vneg) || (0.0 < vneg)) {
    return false;
  }
  if ((0.0 > vpos) || (1.0 < vpos)) {
    return false;
  }

  double vaux = vpos + vneg;
  if ((vaux < this->m_thrcmb) && (vaux > (-1.0 * this->m_thrcmb))) {
    vaux = 0.0;
  }

  combs[2] = (0.0 < vaux) ? vpos : ((0.0 > vaux) ? vneg : 0.0);
  return true;
}

bool Enduce::Dinorms::cmbType1(double *combs) {
  if (!combs) {
    return false;
  }
  double vneg = combs[0];
  double vpos = combs[1];
  if ((-1.0 > vneg) || (0.0 < vneg)) {
    return false;
  }
  if ((0.0 > vpos) || (1.0 < vpos)) {
    return false;
  }

  double vaux = vpos + vneg;
  if ((vaux < this->m_thrcmb) && (vaux > (-1.0 * this->m_thrcmb))) {
    vaux = 0.0;
  }

  combs[2] = (0.0 < vaux) ? (vaux / (1.0 + vneg)) : ((0.0 > vaux) ? (vaux / (1.0 - vpos)) : 0.0);
  return true;
}

bool Enduce::Dinorms::cmbType2(double *combs) {
  if (!combs) {
    return false;
  }
  double vneg = combs[0];
  double vpos = combs[1];
  if ((-1.0 > vneg) || (0.0 < vneg)) {
    return false;
  }
  if ((0.0 > vpos) || (1.0 < vpos)) {
    return false;
  }

  double vaux = vpos + vneg;
  if ((vaux < this->m_thrcmb) && (vaux > (-1.0 * this->m_thrcmb))) {
    vaux = 0.0;
  }

  combs[2] = vaux;
  return true;
}

bool Enduce::Dinorms::cmbType3(double *combs) {
  if (!combs) {
    return false;
  }
  double vneg = combs[0];
  double vpos = combs[1];
  if ((-1.0 > vneg) || (0.0 < vneg)) {
    return false;
  }
  if ((0.0 > vpos) || (1.0 < vpos)) {
    return false;
  }

  double vmul = 0.0;
  double vaux = vpos + vneg;
  if ((vaux < this->m_thrcmb) && (vaux > (-1.0 * this->m_thrcmb))) {
    vaux = 0.0;
  }
  else {
    vmul = 1.0 - (vpos * vneg);
  }

  combs[2] = (0.0 != vaux) ? (vaux * vmul) : 0.0;
  return true;
}



